import {typeOf, Type, bootstrapTypes} from './Type.js'

import {Implementations, isPlainObject, onType} from './helpers.js'
import {alwaysMatches, Any, Multiple} from './TypePatterns.js'

// helper that organizes a list into the same chunks as a pattern is
const collectLike = (list, pattern, state = {pos: 0}) => {
   const result = []
   for (const elt of pattern) {
      if (Array.isArray(elt)) result.push(collectLike(list, elt, state))
      else result.push(list[state.pos++])
   }
   return result
}

export class TypeDispatcher {
   constructor(...specs) {
      this._implementations = {} // maps key to list of [pattern, result] pairs
      this._dependencies = {} // maps key to a map from type vectors to...
      // ...a set of [key, types] that depend on it.
      this._behaviors = {} // maps key to a map from type vectors to results
      this._fallbacks = {} // maps key to a catchall result
      this.merge(bootstrapTypes) // bootstrap the instance
      for (const spec of specs) this.merge(spec)
   }

   // Only modify a TypeDispatcher via merge! Otherwise dependencies will
   // not necessarily be updated properly. Actually I may have it set up so
   // that's the only way to modify a TypeDispatcher... not sure, should
   // probably test it.
   merge(spec) {
      if (!spec) return
      if (typeof spec != 'object') {
         throw new TypeError(
            `TypeDispatcher specifications must be objects, not '${spec}'.`)
      }
      for (const key in spec) {
         let val = spec[key]
         if (val instanceof Type) {
            // TODO: Need to wipe out any dependencies on types[key]!
            this.types[key] = val
            continue
         }
         if (typeof val === 'function') {
            val = onType(Multiple(Any), val)
         }
         if (val instanceof Implementations) {
            if (!(key in this)) {
               // need to set up the item
               Object.defineProperty(this, key, {
                  enumerable: true,
                  configurable: true,
                  get: () => {
                     let tryValue
                     let tryTypes = []
                     try {
                        tryValue = this.resolve(key, [])
                     } catch {
                        // Has no value for the empty type list, so
                        // find a type list it will have a value for
                        tryTypes =
                           this._implementations[key][0][0].sampleTypes()
                        tryValue = this.resolve(key, tryTypes)
                     }
                     // Redefine the property according to what sort of
                     // entity it is:
                     if (typeof tryValue === 'function') {
                        // the usual case: a method of the dispatcher
                        const standard = (...args) => {
                           const types = args.map(typeOf)
                           return this.resolve(key, types)(...args)
                        }
                        standard.resolve = (types) => this.resolve(key, types)
                        standard.isDispatcher = true
                        Object.defineProperty(this, key, {
                           enumerable: true,
                           configurable: true,
                           get: () => standard
                        })
                        return standard
                     }
                     if (tryTypes.length) tryValue = undefined
                     if (typeof tryValue === 'object') {
                        if (!('resolve' in tryValue)) {
                           tryValue.resolve = types => this.resolve(key, types)
                        }
                        const get = () => {
                           if (this._dependencies[key]?.get([])?.size) {
                              return DependencyWatcher(tryValue, [key], this)
                           }
                           return tryValue
                        }
                        Object.defineProperty(this, key, {
                           enumerable: true,
                           configurable: true,
                           get
                        })
                        return get()
                     }
                     Object.defineProperty(this, key, {
                        enumerable: true,
                        configurable: true,
                        get: () => tryValue
                     })
                     return tryValue
                  }
               })
               this._implementations[key] = []
               this._behaviors[key] = new Map()
               this._dependencies[key] = new Map()
            }
            // Now add all of the patterns of this implementation:
            for (const [pattern, result] of val.patterns) {
               if (alwaysMatches(pattern)) {
                  if (key in this._fallbacks) this._disengageFallback(key)
                  this._fallbacks[key] = result
               } else {
                  this._clearBehaviorsMatching(key, pattern)
                  // if it happens the same pattern is already in the
                  // implementations, remove it
                  const imps = this._implementations[key]
                  const have = imps.findIndex(elt => pattern.equal(elt[0]))
                  if (have >= 0) imps.splice(have, 1)
                  this._implementations[key].unshift([pattern, result])
               }
            }
            continue
         }

         if (isPlainObject(val)) {
            this.merge(val)
            continue
         }

         // install value as a catchall value
         this.merge({[key]: onType(Multiple(Any), val)})
      }
   }

   _addToDeps(key, subkey) {
      const depMap = this._dependencies[key]
      if (!depMap.has(subkey)) depMap.set(subkey, new Set())
      const depSet = depMap.get(subkey)
      for (const pair of this.resolve._genDepsOf) depSet.add(pair)
   }

   resolve(key, types) {
      if (!(key in this)) {
         throw new ReferenceError(`no method or value for key '${key}'`)
      }

      if (this.resolve._genDepsOf?.length) _addToDeps(key, types)

      const behave = this._behaviors[key]
      if (behave.has(types)) return behave.get(types)
      const imps = this._implementations[key]
      let needItem = true
      let item
      let pattern
      let template
      if (imps.length) {
         for ([pattern, item] of imps) {
            let finalIndex
            ;[finalIndex, template] = pattern.match(types, 0)
            if (finalIndex === types.length) {
               needItem = false
               break
            }
         }
      }
      if (needItem && key in this._fallbacks) {
         needItem = false
         item = this._fallbacks[key]
         template = [types]
      }
      if (needItem) {
         throw new TypeError(`no matching definition of '${key}' on '${types}'`)
      }
      if (typeof item !== 'function' || 'returns' in item) {
         behave.set(types, item)
         return item
      }
      // item is a Factory. We have to use it to build the behavior
      // First set up to record dependencies
      if (!('_genDepsOf' in this.resolve)) {
         this.resolve._genDepsOf = []
      }
      this.resolve._genDepsOf.push([key, types])
      let theBehavior = item(DependencyRecorder(this, [], this), pattern)
      this.resolve._genDepsOf.pop()
      if (typeof theBehavior === 'function'
          && theBehavior.length
          && template.some(elt => Array.isArray(elt))
      ) {
         // have to wrap the behavior to collect the actual arguments
         // in the way corresponding to the template
         const wrappedBehavior = (...args) => {
            const collectedArgs = collectLike(args, template)
            return theBehavior(...collectedArgs)
         }
         theBehavior = wrappedBehavior
      }
      behave.set(types, theBehavior)
      return theBehavior
   }

   // Method called to invalidate a set of behaviors
   // I think all it has to do is throw them out; they should be
   // regenerated properly by resolve, if all is well.
   _invalidate(behaveSet) {
      for (const [key, types] of behaveSet) this._behaviors[key].delete(types)
   }

   _disengageFallback(key) {
      // We need to find all of the behaviors that currently rely on the
      // fallback, invalidate their dependencies, and remove them.
      const fallTypes = []
      const behs = this._behaviors[key]
      const imps = this._implementations[key]
      const deps = this._dependencies[key]
      for (const types of behs.keys()) {
         let fallsback = true
         for (const [pattern] of imps) {
            const [finalIndex] = pattern.match(types)
            if (finalIndex === types.length) {
               fallsback = false
               break
            }
         }
         if (fallsback) fallTypes.push(types)
      }
      for (const types of fallTypes) {
         const depSet = deps?.get(types)
         if (depSet?.size) this._invalidate(depSet)
         behs.delete(types)
      }
   }

   _clearBehaviorsMatching(key, pattern) {
      // like disengageFallback, just we have the offending pattern:
      const behs = this._behaviors[key]
      const deps = this._dependencies[key]
      const patTypes = behs.keys().filter(types => {
         const [finalIndex] = pattern.match(types)
         return finalIndex === types.length
      })
      for (const types of patTypes) {
         const depSet = deps?.get(types)
         if (depSet?.size) this._invalidate(depSet)
         behs.delete(types)
      }
   }
}

// Proxy that traps accesses and records dependencies on them
const DependencyRecorder = (object, path, repo) => new Proxy(object, {
   get(target, prop, receiver) {
      const result = Reflect.get(target, prop, receiver)
      // pass resolve calls through, since we record dependencies therein:
      if (prop === 'resolve' || 'isDispatcher' in result) return result

      // OK, it's not a method on a TypeDispatcher, it's some other kind of
      // value. So first record the dependency on prop at this path:
      const newPath = path.slice()
      newPath.push(prop)
      const key = newPath[0]
      const subpath = newPath.slice(1)
      repo._addToDeps(key, subpath)
      // Now, if the result is an object, we may need to record further
      // dependencies on its properties (e.g. math.config.predictable)
      // So proxy the return value:
      if (typeof result === 'object') {
         return DependencyRecorder(result, newPath, repo)
      } else return result
   }
})
      
// The flip side: proxy that traps setting properties and invalidates things
// that depend on them:
const DependencyWatcher = (object, path, repo) => new Proxy(object, {
   set(target, prop, value, receiver) {
      // First see if this setting has any dependencies:
      const newPath = path.slice()
      newPath.push(prop)
      const key = newPath.unshift()
      const depSet = repo._dependencies[key]?.get(newPath)
      if (depSet?.size) {
         // It does. So if we are changing it, invalidate them:
         const oldValue = Reflect.get(target, prop, receiver)
         if (value !== oldValue) repo._invalidate(depSet)
      }
      // Now we can just perform the setting
      return Reflect.set(target, prop, value, receiver)
   },
   get(target, prop, receiver) {
      // Only thing we need to do is push the watching down
      const result = Reflect.get(target, prop, receiver)
      if (typeof result === 'object') {
         const newPath = path.slice()
         newPath.push(prop)
         return DependencyWatcher(result, newPath, repo)
      }
      return result
   }
})
