import {Type, Undefined} from './Type.js'

class TypePattern {
   match(typeSequence, position) {
      throw new Error('Specific TypePatterns must implement match')
   }
   sampleTypes() {
      throw new Error('Specific TypePatterns must implement sampleTypes')
   }
   equal(other) {return other.constructor === this.constructor}
}

class MatchTypePattern extends TypePattern {
   constructor(typeToMatch) {
      super()
      this.type = typeToMatch
   }
   match(typeSequence, position) {
      if (position < typeSequence.length
         && typeSequence[position] === this.type
      ) {
         return [position + 1, typeSequence[position]]
      }
      return [-1, Undefined]
   }
   sampleTypes() {return [this.type]}
   equal(other) {return super.equal(other) && this.type === other.type}
}

class SequencePattern extends TypePattern {
   constructor(itemsToMatch) {
      super()
      this.patterns = itemsToMatch.map(pattern)
   }
   match(typeSequence, position) {
      const matches = []
      for (const pat of this.patterns) {
         const [newPos, newMatch] = pat.match(typeSequence, position)
         if (newPos < 0) return [-1, Undefined]
         position = newPos
         matches.push(newMatch)
      }
      return [position, matches]
   }
   sampleTypes() {
      return this.patterns.map(pat => pat.sampleTypes()).flat()
   }
   equal(other) {
      return super.equal(other)
         && this.patterns.length === other.patterns.length
         && this.patterns.every((elt, ix) => elt.equal(other.patterns[ix]))
   }
}

export const pattern = patternOrSpec => {
   if (patternOrSpec instanceof TypePattern) return patternOrSpec
   if (patternOrSpec instanceof Type) {
      return new MatchTypePattern(patternOrSpec)
   }
   if (Array.isArray(patternOrSpec)) return new SequencePattern(patternOrSpec)
   throw new TypeError(`Can't interpret '${patternOrSpec}' as a type pattern`)
}

class AnyPattern extends TypePattern {
   match(typeSequence, position) {
      return position < typeSequence.length
         ? [position + 1, typeSequence[position]]
         : [-1, Undefined]
   }
   sampleTypes() {return [Undefined]}
}

export const Any = new AnyPattern()

class OptionalPattern extends TypePattern {
   constructor(item) {
      this.pattern = pattern(item)
   }
   match(typeSequence, position) {
      const matches = []
      const [newPos, newMatch] = this.pattern.match(typeSequence, position)
      if (newPos >= 0) {
         position = newPos
         matches.push(newMatch)
      }
      return [position, matches]
   }
   sampleTypes() {return []}
   equal(other) {
      return super.equal(other) && this.pattern.equal(other.pattern)
   }
}

export const Optional = item => new OptionalPattern(item)

class MultiPattern extends TypePattern {
   constructor(item) {
      super()
      this.pattern = pattern(item)
   }
   match(typeSequence, position) {
      const matches = []
      while (true) {
         const [newPos, newMatch] = this.pattern.match(typeSequence, position)
         if (newPos < 0) return [position, matches]
         position = newPos
         matches.push(newMatch)
      }
   }
   sampleTypes() {return []}
   equal(other) {
      return super.equal(other) && this.pattern.equal(other.pattern)
   }
}

export const Multiple = item => new MultiPattern(item)
export const alwaysMatches =
   pat => pat instanceof MultiPattern && pat.pattern instanceof AnyPattern
